@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: CDDL-1.0
 /*
  * CDDL HEADER START
  *
@@ -6,7 +7,7 @@
  * You may not use this file except in compliance with the License.
  *
  * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
- * or http://www.opensolaris.org/os/licensing.
+ * or https://opensource.org/licenses/CDDL-1.0.
  * See the License for the specific language governing permissions
  * and limitations under the License.
  *
@@ -28,6 +29,14 @@
 
 #include <linux/compiler.h>
 
+#if !defined(zfs_fallthrough)
+#if defined(HAVE_IMPLICIT_FALLTHROUGH)
+#define	zfs_fallthrough		__attribute__((__fallthrough__))
+#else
+#define	zfs_fallthrough		((void)0)
+#endif
+#endif
+
 #if !defined(READ_ONCE)
 #define	READ_ONCE(x)		ACCESS_ONCE(x)
 #endif
