@@ -1236,12 +1236,21 @@ vdev_geom_io_done(zio_t *zio)
 	struct bio *bp = zio->io_bio;
 
 	if (zio->io_type != ZIO_TYPE_READ && zio->io_type != ZIO_TYPE_WRITE) {
-		ASSERT3P(bp, ==, NULL);
+		ASSERT0P(bp);
 		return;
 	}
 
 	if (bp == NULL) {
-		ASSERT3S(zio->io_error, ==, ENXIO);
+		if (zio_injection_enabled && zio->io_error == EIO)
+			/*
+			 * Convert an injected EIO to ENXIO. This is needed to
+			 * work around zio_handle_device_injection_impl() not
+			 * currently being able to inject ENXIO directly, while
+			 * the assertion below only allows ENXIO here.
+			 */
+			zio->io_error = SET_ERROR(ENXIO);
+		else
+			ASSERT3S(zio->io_error, ==, ENXIO);
 		return;
 	}
 
@@ -1276,7 +1285,8 @@ vdev_ops_t vdev_disk_ops = {
 	.vdev_op_fini = NULL,
 	.vdev_op_open = vdev_geom_open,
 	.vdev_op_close = vdev_geom_close,
-	.vdev_op_asize = vdev_default_asize,
+	.vdev_op_psize_to_asize = vdev_default_asize,
+	.vdev_op_asize_to_psize = vdev_default_psize,
 	.vdev_op_min_asize = vdev_default_min_asize,
 	.vdev_op_min_alloc = NULL,
 	.vdev_op_io_start = vdev_geom_io_start,
