@@ -302,13 +302,8 @@ spl_kmem_free_impl(const void *buf, size_t size)
 #ifdef DEBUG_KMEM
 
 /* Shim layer memory accounting */
-#ifdef HAVE_ATOMIC64_T
 atomic64_t kmem_alloc_used = ATOMIC64_INIT(0);
-unsigned long long kmem_alloc_max = 0;
-#else  /* HAVE_ATOMIC64_T */
-atomic_t kmem_alloc_used = ATOMIC_INIT(0);
-unsigned long long kmem_alloc_max = 0;
-#endif /* HAVE_ATOMIC64_T */
+uint64_t kmem_alloc_max = 0;
 
 EXPORT_SYMBOL(kmem_alloc_used);
 EXPORT_SYMBOL(kmem_alloc_max);
@@ -320,9 +315,9 @@ spl_kmem_alloc_debug(size_t size, int flags, int node)
 
 	ptr = spl_kmem_alloc_impl(size, flags, node);
 	if (ptr) {
-		kmem_alloc_used_add(size);
-		if (unlikely(kmem_alloc_used_read() > kmem_alloc_max))
-			kmem_alloc_max = kmem_alloc_used_read();
+		atomic64_add(size, &kmem_alloc_used);
+		if (unlikely(atomic64_read(&kmem_alloc_used) > kmem_alloc_max))
+			kmem_alloc_max = atomic64_read(&kmem_alloc_used);
 	}
 
 	return (ptr);
@@ -331,7 +326,7 @@ spl_kmem_alloc_debug(size_t size, int flags, int node)
 inline void
 spl_kmem_free_debug(const void *ptr, size_t size)
 {
-	kmem_alloc_used_sub(size);
+	atomic64_sub(size, &kmem_alloc_used);
 	spl_kmem_free_impl(ptr, size);
 }
 
@@ -595,7 +590,7 @@ spl_kmem_init(void)
 {
 
 #ifdef DEBUG_KMEM
-	kmem_alloc_used_set(0);
+	atomic64_set(&kmem_alloc_used, 0);
 
 
 
@@ -617,9 +612,10 @@ spl_kmem_fini(void)
 	 * at that address to aid in debugging.  Performance is not
 	 * a serious concern here since it is module unload time.
 	 */
-	if (kmem_alloc_used_read() != 0)
+	if (atomic64_read(&kmem_alloc_used) != 0)
 		printk(KERN_WARNING "kmem leaked %ld/%llu bytes\n",
-		    (unsigned long)kmem_alloc_used_read(), kmem_alloc_max);
+		    (unsigned long)atomic64_read(&kmem_alloc_used),
+		    kmem_alloc_max);
 
 #ifdef DEBUG_KMEM_TRACKING
 	spl_kmem_fini_tracking(&kmem_list, &kmem_lock);
