@@ -160,9 +160,31 @@ zfs_userquota_prop_to_obj(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type)
 	}
 }
 
+static uint64_t
+zfs_usedquota_prop_to_default(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type)
+{
+	switch (type) {
+	case ZFS_PROP_USERUSED:
+		return (zfsvfs->z_defaultuserquota);
+	case ZFS_PROP_USEROBJUSED:
+		return (zfsvfs->z_defaultuserobjquota);
+	case ZFS_PROP_GROUPUSED:
+		return (zfsvfs->z_defaultgroupquota);
+	case ZFS_PROP_GROUPOBJUSED:
+		return (zfsvfs->z_defaultgroupobjquota);
+	case ZFS_PROP_PROJECTUSED:
+		return (zfsvfs->z_defaultprojectquota);
+	case ZFS_PROP_PROJECTOBJUSED:
+		return (zfsvfs->z_defaultprojectobjquota);
+	default:
+		return (0);
+	}
+}
+
 int
 zfs_userspace_many(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type,
-    uint64_t *cookiep, void *vbuf, uint64_t *bufsizep)
+    uint64_t *cookiep, void *vbuf, uint64_t *bufsizep,
+    uint64_t *default_quota)
 {
 	int error;
 	zap_cursor_t zc;
@@ -187,6 +209,8 @@ zfs_userspace_many(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type,
 	    !dmu_objset_userobjspace_present(zfsvfs->z_os))
 		return (SET_ERROR(ENOTSUP));
 
+	*default_quota = zfs_usedquota_prop_to_default(zfsvfs, type);
+
 	obj = zfs_userquota_prop_to_obj(zfsvfs, type);
 	if (obj == ZFS_NO_OBJECT) {
 		*bufsizep = 0;
@@ -350,7 +374,7 @@ zfs_set_userquota(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type,
 	if (*objp == 0) {
 		*objp = zap_create(zfsvfs->z_os, DMU_OT_USERGROUP_QUOTA,
 		    DMU_OT_NONE, 0, tx);
-		VERIFY(0 == zap_add(zfsvfs->z_os, MASTER_NODE_OBJ,
+		VERIFY0(zap_add(zfsvfs->z_os, MASTER_NODE_OBJ,
 		    zfs_userquota_prop_prefixes[type], 8, 1, objp, tx));
 	}
 	mutex_exit(&zfsvfs->z_lock);
@@ -362,7 +386,7 @@ zfs_set_userquota(zfsvfs_t *zfsvfs, zfs_userquota_prop_t type,
 	} else {
 		err = zap_update(zfsvfs->z_os, *objp, buf, 8, 1, &quota, tx);
 	}
-	ASSERT(err == 0);
+	ASSERT0(err);
 	if (fuid_dirtied)
 		zfs_fuid_sync(zfsvfs, tx);
 	dmu_tx_commit(tx);
@@ -373,7 +397,7 @@ boolean_t
 zfs_id_overobjquota(zfsvfs_t *zfsvfs, uint64_t usedobj, uint64_t id)
 {
 	char buf[20 + DMU_OBJACCT_PREFIX_LEN];
-	uint64_t used, quota, quotaobj;
+	uint64_t used, quota, quotaobj, default_quota = 0;
 	int err;
 
 	if (!dmu_objset_userobjspace_present(zfsvfs->z_os)) {
@@ -399,20 +423,29 @@ zfs_id_overobjquota(zfsvfs_t *zfsvfs, uint64_t usedobj, uint64_t id)
 			return (B_FALSE);
 		}
 		quotaobj = zfsvfs->z_projectobjquota_obj;
+		default_quota = zfsvfs->z_defaultprojectobjquota;
 	} else if (usedobj == DMU_USERUSED_OBJECT) {
 		quotaobj = zfsvfs->z_userobjquota_obj;
+		default_quota = zfsvfs->z_defaultuserobjquota;
 	} else if (usedobj == DMU_GROUPUSED_OBJECT) {
 		quotaobj = zfsvfs->z_groupobjquota_obj;
+		default_quota = zfsvfs->z_defaultgroupobjquota;
 	} else {
 		return (B_FALSE);
 	}
-	if (quotaobj == 0 || zfsvfs->z_replay)
+	if (zfsvfs->z_replay)
 		return (B_FALSE);
 
 	(void) snprintf(buf, sizeof (buf), "%llx", (longlong_t)id);
-	err = zap_lookup(zfsvfs->z_os, quotaobj, buf, 8, 1, &quota);
-	if (err != 0)
-		return (B_FALSE);
+	if (quotaobj == 0) {
+		if (default_quota == 0)
+			return (B_FALSE);
+		quota = default_quota;
+	} else {
+		err = zap_lookup(zfsvfs->z_os, quotaobj, buf, 8, 1, &quota);
+		if (err != 0 && ((quota = default_quota) == 0))
+			return (B_FALSE);
+	}
 
 	(void) snprintf(buf, sizeof (buf), DMU_OBJACCT_PREFIX "%llx",
 	    (longlong_t)id);
@@ -426,7 +459,7 @@ boolean_t
 zfs_id_overblockquota(zfsvfs_t *zfsvfs, uint64_t usedobj, uint64_t id)
 {
 	char buf[20];
-	uint64_t used, quota, quotaobj;
+	uint64_t used, quota, quotaobj, default_quota = 0;
 	int err;
 
 	if (usedobj == DMU_PROJECTUSED_OBJECT) {
@@ -441,20 +474,29 @@ zfs_id_overblockquota(zfsvfs_t *zfsvfs, uint64_t usedobj, uint64_t id)
 			return (B_FALSE);
 		}
 		quotaobj = zfsvfs->z_projectquota_obj;
+		default_quota = zfsvfs->z_defaultprojectquota;
 	} else if (usedobj == DMU_USERUSED_OBJECT) {
 		quotaobj = zfsvfs->z_userquota_obj;
+		default_quota = zfsvfs->z_defaultuserquota;
 	} else if (usedobj == DMU_GROUPUSED_OBJECT) {
 		quotaobj = zfsvfs->z_groupquota_obj;
+		default_quota = zfsvfs->z_defaultgroupquota;
 	} else {
 		return (B_FALSE);
 	}
-	if (quotaobj == 0 || zfsvfs->z_replay)
+	if (zfsvfs->z_replay)
 		return (B_FALSE);
 
 	(void) snprintf(buf, sizeof (buf), "%llx", (longlong_t)id);
-	err = zap_lookup(zfsvfs->z_os, quotaobj, buf, 8, 1, &quota);
-	if (err != 0)
-		return (B_FALSE);
+	if (quotaobj == 0) {
+		if (default_quota == 0)
+			return (B_FALSE);
+		quota = default_quota;
+	} else {
+		err = zap_lookup(zfsvfs->z_os, quotaobj, buf, 8, 1, &quota);
+		if (err != 0 && ((quota = default_quota) == 0))
+			return (B_FALSE);
+	}
 
 	err = zap_lookup(zfsvfs->z_os, usedobj, buf, 8, 1, &used);
 	if (err != 0)
